/*
 * Copyright (c) @ Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 	This product includes software developed by the MASH Research
 * 	Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * 
 */

#include "net.h"
#include "rtp.h"
#include "timer.h"
#include "session-rtp.h"
#include "transmitter-rtp.h"
#include "source.h"
#include "rtp-record.h"


class RTP_SessionArchiveAudio : public AudioSession {
public:
	RTP_SessionArchiveAudio() : AudioSession() {
		
	}
	~RTP_SessionArchiveAudio() {}
	virtual void recv(CtrlHandler *ch);
};

class RTP_SessionArchiveVideo : public VideoSession {
public:
	RTP_SessionArchiveVideo() : VideoSession()  {
		
	}
	~RTP_SessionArchiveVideo() {}
	virtual void recv(CtrlHandler *ch);
};


static class ArchiveAudioSessionClass : public TclClass {
    public:
	ArchiveAudioSessionClass() : TclClass("Session/RTP/Audio/Archive") {}
	TclObject* create(int /*argc*/, const char*const* /*argv*/) {
		return (new  RTP_SessionArchiveAudio);
	}
} archive_audio_class;


static class ArchiveVideoSessionClass : public TclClass {
    public:
	ArchiveVideoSessionClass() : TclClass("Session/RTP/Video/Archive") {}
	TclObject* create(int /*argc*/, const char*const* /*argv*/) {
		return (new RTP_SessionArchiveVideo);
	}
} archive_video_class;




void RTP_SessionArchiveAudio::recv(CtrlHandler* ch)
{
	u_int32_t src;
	int srcport;
	pktbuf *my = (pktbuf *)pktbuf_;
	int cc = ch->recv(my->data, 2 * RTP_MTU, src, srcport);

	MTrace(trcArchive, ("session cc=%d", cc));
	if (cc <= 0)
		return;
	
	rtcphdr* rh = (rtcphdr*)(my->data);
	if (cc < int(sizeof(*rh))) {
		++nrunt_;
		return;
	}
	/*
	 * try to filter out junk: first thing in packet must be
	 * sr, rr or bye & version number must be correct.
	 */
	switch(ntohs(rh->rh_flags) & 0xc0ff) {
	case RTP_VERSION << 14 | RTCP_PT_SR:
	case RTP_VERSION << 14 | RTCP_PT_RR:
	case RTP_VERSION << 14 | RTCP_PT_BYE:
		break;
	default:
		/*
		 * XXX should further categorize this error -- it is
		 * likely that people mis-implement applications that
		 * don't put something other than SR,RR,BYE first.
		 */
		++badversion_;
		return;
	}
	/*
	 * at this point we think the packet's valid.  Update our average
	 * size estimator.  Also, there's valid ssrc so charge errors to it
	 */
	ch->sample_size(cc);
	u_int32_t addr = src;

	/*
	 * First record in compound packet must be the ssrc of the
	 * sender of the packet.  Pull it out here so we can use
	 * it in the sdes parsing, since the sdes record doesn't
	 * contain the ssrc of the sender (in the case of mixers).
	 */
	u_int32_t ssrc = rh->rh_ssrc;
	Source* ps = sm_->lookup(this, ssrc, ssrc, addr);
	if (ps == 0)
		return;

	int layer = ch - ch_;
	/*
	 * Outer loop parses multiple RTCP records of a "compound packet".
	 * There is no framing between records.  Boundaries are implicit
	 * and the overall length comes from UDP.
	 */
	u_char* epack = (u_char*)rh + cc;
	while ((u_char*)rh < epack) {
		u_int len = (ntohs(rh->rh_len) << 2) + 4;
		u_char* ep = (u_char*)rh + len;
		if (ep > epack) {
			ps->badsesslen(1);
			return;
		}
		u_int flags = ntohs(rh->rh_flags);
		if (flags >> 14 != RTP_VERSION) {
			ps->badsessver(1);
			return;
		}
		switch (flags & 0xff) {

		case RTCP_PT_SR:
			parse_sr(rh, flags, ep, ps, addr, layer);
			break;

		case RTCP_PT_RR:
			parse_rr(rh, flags, ep, ps, addr, layer);
			break;

		case RTCP_PT_SDES:
			parse_sdes(rh, flags, ep, ps, addr, ssrc, layer);
			break;

		case RTCP_PT_BYE:
			parse_bye(rh, flags, ep, ps);
			break;

		case RTCP_PT_APP:
			parse_app(rh, flags, ep, ps);
			break;

		default:
			ps->badsessopt(1);
			break;
		}
		rh = (rtcphdr*)ep;
	}
	
	PacketModule* h = ps->ctrl_handler();
	if (h==0)
		return;
	// For now, doesn't record any ctrl packets until data starts.  
	// Need to change this to record recv reports.
	((pktbuf *)pktbuf_)->layer=layer;
	((pktbuf *)pktbuf_)->len = cc;
	h->recv((pktbuf *)pktbuf_);
	
}

//Note: audio and video have different policies right now with regards
//to recording listeners.  It should default to not recording them, and
//have an option to allow it.

void RTP_SessionArchiveVideo::recv(CtrlHandler* ch)
{
	u_int32_t src;
	int srcport;
	pktbuf *my = (pktbuf *)pktbuf_;
	int cc = ch->recv(my->data, 2 * RTP_MTU, src, srcport);
	if (cc <= 0)
		return;

	rtcphdr* rh = (rtcphdr*)(my->data);
	if (cc < int(sizeof(*rh))) {
		++nrunt_;
		return;
	}
	/*
	 * try to filter out junk: first thing in packet must be
	 * sr, rr or bye & version number must be correct.
	 */
	switch(ntohs(rh->rh_flags) & 0xc0ff) {
	case RTP_VERSION << 14 | RTCP_PT_SR:
	case RTP_VERSION << 14 | RTCP_PT_RR:
	case RTP_VERSION << 14 | RTCP_PT_BYE:
		break;
	default:
		/*
		 * XXX should further categorize this error -- it is
		 * likely that people mis-implement applications that
		 * don't put something other than SR,RR,BYE first.
		 */
		++badversion_;
		return;
	}
	/*
	 * at this point we think the packet's valid.  Update our average
	 * size estimator.  Also, there's valid ssrc so charge errors to it
	 */
	ch->sample_size(cc);
	u_int32_t addr = src;

	/*
	 * First record in compound packet must be the ssrc of the
	 * sender of the packet.  Pull it out here so we can use
	 * it in the sdes parsing, since the sdes record doesn't
	 * contain the ssrc of the sender (in the case of mixers).
	 */
	u_int32_t ssrc = rh->rh_ssrc;
	Source* ps = sm_->lookup(this, ssrc, ssrc, addr);
	if (ps == 0)
		return;

	int layer = ch - ch_;
	
	/*
	 * Outer loop parses multiple RTCP records of a "compound packet".
	 * There is no framing between records.  Boundaries are implicit
	 * and the overall length comes from UDP.
	 */
	u_char* epack = (u_char*)rh + cc;
	//printf("session-record cc=%d pb->len=%d\n", cc, rh->rh_len);

	
		
	while ((u_char*)rh < epack) {
		u_int len = (ntohs(rh->rh_len) << 2) + 4;
		u_char* ep = (u_char*)rh + len;
		if (ep > epack) {
			ps->badsesslen(1);
			return;
		}
		u_int flags = ntohs(rh->rh_flags);
		if (flags >> 14 != RTP_VERSION) {
			ps->badsessver(1);
			return;
		}
		//printf("sess %d\n", flags & 0xff);
		switch (flags & 0xff) {

		case RTCP_PT_SR:
			parse_sr(rh, flags, ep, ps, addr, layer);
			break;

		case RTCP_PT_RR:
			parse_rr(rh, flags, ep, ps, addr, layer);
			break;

		case RTCP_PT_SDES:
			parse_sdes(rh, flags, ep, ps, addr, ssrc, layer);
			break;

		case RTCP_PT_BYE:
			parse_bye(rh, flags, ep, ps);
			break;

		case RTCP_PT_APP:
			parse_app(rh, flags, ep, ps);
			break;

		default:
			ps->badsessopt(1);
			break;
		}
		rh = (rtcphdr*)ep;
	}
	
	PacketModule* h = ps->ctrl_handler();
	if (h==0) {
		// For now, doesn't record any ctrl packets until data starts.  
		// Need to change this to record recv reports.
		//h = ps->activate(-1);  // What should the format be?
		//h = ps->ctrl_handler();
		return;
	}
	((pktbuf *)pktbuf_)->layer=layer;
	((pktbuf *)pktbuf_)->len = cc;
	h->recv((pktbuf *)pktbuf_);
}

